/* Hard disk drive temperature plugin for Gkrellm
 * Copyright (C) 2001  Emmanuel VARAGNAT <coredump@free.fr>

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>
#include <stdio.h>
#include <fcntl.h>
#include <malloc.h>
#include <string.h>
#include <wait.h>

#include <gkrellm/gkrellm.h>

#define	CONFIG_NAME	"HDDtemp"
#define	STYLE_NAME	"hddtemp"

typedef struct {
  TextStyle	ts;
  gint		lbearing,
                rbearing,
                width,
                ascent,
                descent;
} Extents;

typedef struct {
  char   *name;
  Panel	 *panel;
  Decal  *d_hdd, *d_temp;
} Drive;
  

static Extents  hdd_extents, temp_extents;
static Monitor	*mon_hddtemp;
static gint	style_id;
static Drive    **drives = NULL;
static guint    nb_drives;

static void string_extents(gchar *string, Extents *ext) {
  gdk_string_extents(ext->ts.font, string, &ext->lbearing, &ext->rbearing,
		     &ext->width, &ext->ascent, &ext->descent);
}

static int drive_compare(const void* s1, const void*s2) {
  return strcmp((*(Drive**)s1)->name, (*(Drive**)s2)->name);
}

static char *get_hdtemp(const char *drive) {
  int old0, old1, old2;
  int fd[2];
  pid_t pid;

  old0 = dup(0);
  old1 = dup(1);
  close(0);
  close(1);

  if(pipe(fd) < 0) {
    dup2(old0, 0);
    dup2(old1, 1);
    perror("pipe");
    exit(1);
  }
  
  switch((pid = fork())) {
  case 0:
    {
      char dev[128];
      char *args[3]= {
	"hddtemp", "/dev/XXX", NULL
      };

      close(0);
      old2 = dup(2);
      close(2);
      
      snprintf(dev, sizeof(dev), "/dev/%s", drive);
      args[1] = dev;
      execvp(args[0], args);

      /* re-open standard input and outputs */
      dup2(old0, 0);
      dup2(old1, 1);
      dup2(old2, 2);
      
      fprintf(stderr, "Can't execute hddtemp\n");
      exit(1);
      break;
    }
    break;
  case -1:
    dup2(old0, 0);
    dup2(old1, 1);
    perror("fork");
    exit(1);
    break;
  default:
    {
      char ch, line[255], *p = NULL;
      unsigned int n = 0, nb_line = 0;

      close(1);
      
      while(read(fd[0], &ch, 1)) {
	if(n < 255 && ch == '\n') {
	  line[n] = '\0';
	  p = strchr(line,'');
	  *(p+2) = '\0';
	  p--;
	  while(p != line && (*p >= '0' && *p <= '9')) {
	    p--;
	  }

	  n = 0;
	  nb_line++;
	}
	else {
	  line[n++] = ch;
	}
      }
      
      close(fd[0]);

      /* rouvre la sortie et l'entre standard */
      dup2(old0, 0);
      dup2(old1, 1);

      waitpid(pid, NULL, 0);

      if(!n && !nb_line)
	return NULL;
      
      return strdup(p+1);
    }
  }
  return NULL;
}

static void scan_drives() {
  unsigned int i, found;
  struct dirent *entry;
  DIR *dir;

  drives = (Drive**)malloc(sizeof(Drive*));
  nb_drives = 1;
  drives[0] = NULL;
  found = 0;

  dir = opendir("/proc/ide");
  if(!dir) {
    perror("opendir");
    exit(1);
  }

  while((entry = readdir(dir))) {
    if(entry->d_name[0] == 'h' && entry->d_name[1] == 'd') {
      if(found+1 > nb_drives) {
	drives = (Drive**) realloc(drives ,++nb_drives*sizeof(Drive*));
	if(drives == NULL) {
	  perror("realloc");
	  exit(1);
	}
      }

      drives[found] = (Drive*) malloc(sizeof(Drive));
      drives[found++]->name = strdup(entry->d_name);
    }
  }
  closedir(dir);

  qsort(drives, nb_drives, sizeof(Drive*), drive_compare);

  /* we are scanning one time to eliminate CDROM and such */
  for(i = 0; i < nb_drives; i++) {
    if(drives[i]) {
      char *temp;

      temp = get_hdtemp(drives[i]->name);
      if(temp) {
	free(temp);
      }
      else {
	free(drives[i]);
	drives[i] = NULL;
      }
    }
  }
}

static void update_plugin() {
  /* every minute */
  if ((GK.timer_ticks % 600) == 0) {
    int i;
    char *temp;

    for(i = 0; i < nb_drives; i++) {
      if(drives[i]) {
      	gkrellm_draw_decal_text(drives[i]->panel, drives[i]->d_hdd, drives[i]->name, -1);
	temp = get_hdtemp(drives[i]->name);
	gkrellm_draw_decal_text(drives[i]->panel, drives[i]->d_temp, temp, -1);
	free(temp);

	gkrellm_draw_panel_layers(drives[i]->panel);
      }
    }
  }
}

static gint panel_expose_event(GtkWidget *widget, GdkEventExpose *ev, gpointer p) {
  Drive *drv = (Drive*)p;

  gdk_draw_pixmap(widget->window,
		  widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
		  drv->panel->pixmap,
		  ev->area.x, ev->area.y,
		  ev->area.x, ev->area.y,
		  ev->area.width, ev->area.height);
  return FALSE;
}

static void create_plugin(GtkWidget *vbox, gint first_create) {
  Style	*style;
  int i;
  
  for(i = 0; i < nb_drives; i++) {
    if(drives[i]) {
      if (first_create)
	drives[i]->panel = gkrellm_panel_new0();
      style = gkrellm_meter_style(style_id);
      
      hdd_extents.ts = *gkrellm_meter_textstyle(style_id);
      string_extents("8M", &hdd_extents);
      temp_extents.ts = *gkrellm_meter_alt_textstyle(style_id);
      temp_extents.ts.font = GK.large_font;
      string_extents("88M", &temp_extents);
      
      drives[i]->d_hdd =  gkrellm_create_decal_text(drives[i]->panel, "HDD", &hdd_extents.ts,
						    style, 0, -1, hdd_extents.width + 2);
      drives[i]->d_temp = gkrellm_create_decal_text(drives[i]->panel, "8888", &temp_extents.ts,
						    style, 0, -1, temp_extents.width + 2);
      drives[i]->d_hdd->x = 3;
      drives[i]->d_temp->x = drives[i]->d_hdd->x + drives[i]->d_hdd->w + 3;
      drives[i]->d_temp->y = 4;
      drives[i]->d_hdd->y = drives[i]->d_temp->y + temp_extents.ascent - hdd_extents.ascent;
      
      gkrellm_panel_configure(drives[i]->panel, NULL, style);
      gkrellm_panel_create(vbox, mon_hddtemp, drives[i]->panel);
      
      if (first_create)
	gtk_signal_connect(GTK_OBJECT (drives[i]->panel->drawing_area), "expose_event",
			   (GtkSignalFunc) panel_expose_event, drives[i]);
    }
  }
}

static void config_plugin(GtkWidget *vbox) {
  gtk_box_pack_start(GTK_BOX(vbox), gtk_label_new("Gkrellm plugin for hddtemp (v0.1)\n\n"
						  "contact me: coredump@free.fr\n"
						  "website: http://coredump.free.fr/linux"), TRUE, TRUE, 0);
}

static Monitor	plugin_mon = {
  CONFIG_NAME,       /* Title for config clist.   */
  0,		     /* Id,  0 if a plugin       */
  create_plugin,     /* The create function      */
  update_plugin,     /* The update function      */
  config_plugin,     /* The config tab create function   */
  NULL,		     /* Apply the config function        */
  
  NULL,		     /* Save user config */
  NULL,		     /* Load user config */
  NULL,		     /* config keyword   */
  
  NULL,		      /* Undefined 2	*/
  NULL,		      /* Undefined 1	*/
  NULL,		      /* private	*/
  
  MON_DISK,	      /* Insert plugin before this monitor	*/
  
  NULL,		      /* Handle if a plugin, filled in by GKrellM     */
  NULL		      /* path if a plugin, filled in by GKrellM       */
};


Monitor * init_plugin() {
  scan_drives(); 
  style_id = gkrellm_add_meter_style(&plugin_mon, STYLE_NAME);
  mon_hddtemp = &plugin_mon;
  return &plugin_mon;
}
